import os
from typing import Optional, Callable

from dataclasses import dataclass

from petisco.application.config.get_funtion_from_string import get_function_from_string
from petisco.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)
from petisco.notifier.infrastructure.not_implemented_notifier import (
    NotImplementedNotifier,
)
from petisco.notifier.infrastructure.slack.slack_notifier_message_converter import (
    SlackNotifierMessageConverter,
)
from petisco.notifier.infrastructure.slack.slack_notifier import SlackNotifier


def get_slack_notifier():
    slack_token = os.environ.get("SLACK_API_TOKEN", None)
    slack_channel = os.environ.get("SLACK_CHANNEL", "#feed")
    if not slack_token:
        raise ConnectionError(
            "Petisco Slack Notifier is not configured. "
            "Please check Slack Notifier configuration and add required values"
        )
    return SlackNotifier(
        token=slack_token,
        channel=slack_channel,
        converter=SlackNotifierMessageConverter(),
    )


def get_default_notifier():
    return NotImplementedNotifier()


@dataclass
class ConfigNotifier:
    provider: Optional[Callable] = None

    @staticmethod
    def from_dict(kdict):
        provider = get_default_notifier
        if kdict and isinstance(kdict, dict):
            provider = (
                get_function_from_string(kdict.get("provider"))
                .handle(
                    on_failure=raise_petisco_config_exception,
                    failure_args=(kdict, "notifier:provider"),
                )
                .unwrap()
            )

        return ConfigNotifier(provider=provider)



import pytest

from petisco import LogMessage


@pytest.mark.unit
def test_should_create_a_log_message_and_convert_to_dict(given_any_info_id):

    log_message = LogMessage(
        layer="use_case",
        operation="test",
        info_id=given_any_info_id,
        data={"message": "ok"},
    )

    dict_message = log_message.to_dict()

    assert dict_message == {
        "data": {"message": "ok"},
        "meta": {
            "layer": "use_case",
            "operation": "test",
            "info_id": given_any_info_id.to_dict(),
        },
    }


@pytest.mark.unit
def test_should_create_a_log_message_and_convert_to_dict_without_info_id():

    log_message = LogMessage(layer="use_case", operation="test", data={"message": "ok"})

    dict_message = log_message.to_dict()

    assert dict_message == {
        "data": {"message": "ok"},
        "meta": {"layer": "use_case", "operation": "test"},
    }


@pytest.mark.unit
def test_should_create_a_log_message_and_convert_to_dict_and_add_meta_level():

    log_message = LogMessage(layer="use_case", operation="test", data={"message": "ok"})

    dict_message = log_message.to_dict()

    dict_message["meta"]["level"] = "info"

    assert dict_message == {
        "data": {"message": "ok"},
        "meta": {"level": "info", "layer": "use_case", "operation": "test"},
    }


@pytest.mark.unit
def test_should_create_a_log_message_without_layer():

    log_message = LogMessage(operation="test", data={"message": "ok"})

    dict_message = log_message.to_dict()

    dict_message["meta"]["level"] = "info"

    assert dict_message == {
        "data": {"message": "ok"},
        "meta": {"level": "info", "operation": "test"},
    }


@pytest.mark.unit
def test_should_create_a_log_message_without_operation():

    log_message = LogMessage(layer="use_case", data={"message": "ok"})

    dict_message = log_message.to_dict()

    dict_message["meta"]["level"] = "info"

    assert dict_message == {
        "data": {"message": "ok"},
        "meta": {"level": "info", "layer": "use_case"},
    }



import pytest

from petisco.fixtures.testing_decorators import testing_with_rabbitmq
from tests.modules.event.mothers.event_user_created_mother import EventUserCreatedMother
from tests.modules.event.mothers.rabbitmq_event_bus_mother import RabbitMqEventBusMother
from tests.modules.event.mothers.rabbitmq_event_configurer_mother import (
    RabbitMqEventConfigurerMother,
)


@pytest.mark.integration
@testing_with_rabbitmq
def test_should_publish_event_without_rabbitmq_configuration():
    event = EventUserCreatedMother.random()

    bus = RabbitMqEventBusMother.default()
    bus.publish(event)

    bus.configurer.clear()


@pytest.mark.integration
@testing_with_rabbitmq
def test_should_publish_event_with_previous_rabbitmq_configuration():
    event = EventUserCreatedMother.random()

    configurer = RabbitMqEventConfigurerMother.default()
    configurer.configure_event(event)

    bus = RabbitMqEventBusMother.default()
    bus.publish(event)

    configurer.clear()



import os
import string
import random

import pytest

from petisco import RabbitMqConnector
from petisco.event.configurer.infrastructure.rabbitmq_configurer import (
    RabbitMqEventConfigurer,
)


@pytest.fixture
def given_any_organization() -> str:
    return f"acme"


@pytest.fixture
def given_any_service() -> str:
    return f"petisco"


@pytest.fixture
def given_rabbitmq_connector():
    return RabbitMqConnector()


@pytest.fixture
def given_rabbitmq_configurer(
    given_rabbitmq_connector, given_any_organization, given_any_service
):
    configurer = RabbitMqEventConfigurer(
        given_rabbitmq_connector, given_any_organization, given_any_service
    )
    yield configurer
    configurer.clear()


@pytest.fixture
def given_any_topic():
    return "topic"


@pytest.fixture
def given_random_str() -> str:
    letters = string.ascii_letters
    return "".join(random.choice(letters) for i in range(5)).lower()


@pytest.fixture
def given_random_organization(given_random_str) -> str:
    return f"acme-{given_random_str}"


@pytest.fixture
def given_random_service(given_random_str) -> str:
    return f"service-{given_random_str}"


@pytest.fixture
def given_random_topic(given_random_str) -> str:
    return f"topic-{given_random_str}"


@pytest.fixture
def given_a_short_message_ttl():
    original_value = os.environ.get("RABBITMQ_MESSAGE_TTL")

    os.environ["RABBITMQ_MESSAGE_TTL"] = "10"

    yield

    if original_value:
        os.environ["RABBITMQ_MESSAGE_TTL"] = original_value



import pytest
from meiga import Success, isFailure, Failure

from petisco import controller_handler, ERROR, __version__, DEBUG
from petisco.domain.errors.critical_error import CriticalError
from petisco.event.shared.domain.request_responded import RequestResponded
from tests.modules.unit.mocks.fake_event_publisher import FakeEventPublisher
from tests.modules.unit.mocks.fake_logger import FakeLogger
from tests.modules.unit.mocks.fake_notifier import FakeNotifier
from tests.modules.unit.mocks.log_message_mother import LogMessageMother


@pytest.mark.unit
def test_should_execute_successfully_a_empty_controller_without_input_parameters(
    given_any_petisco, given_any_info_id, given_headers_provider
):

    logger = FakeLogger()
    publisher = FakeEventPublisher()

    @controller_handler(
        app_name="petisco",
        app_version=__version__,
        logger=logger,
        send_request_responded_event=True,
        publisher=publisher,
        headers_provider=given_headers_provider(given_any_info_id.get_http_headers()),
    )
    def my_controller():
        return Success("Hello Petisco")

    http_response = my_controller()

    assert http_response == ({"message": "OK"}, 200)

    first_logging_message = logger.get_logging_messages()[0]
    second_logging_message = logger.get_logging_messages()[1]

    assert first_logging_message == (
        DEBUG,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Processing Request",
            info_id=given_any_info_id,
        ).to_dict(),
    )
    assert second_logging_message == (
        DEBUG,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Result[status: success | value: Hello Petisco]",
            info_id=given_any_info_id,
        ).to_dict(),
    )

    request_responded = publisher.get_sent_events()[0]
    assert isinstance(request_responded, RequestResponded)
    assert request_responded.app_name == "petisco"
    assert request_responded.app_version == __version__
    assert request_responded.controller == "my_controller"
    assert request_responded.is_success is True
    assert request_responded.http_response["content"] == {
        "message": '{"message": "OK"}',
        "message_size": 17,
    }
    assert request_responded.http_response["status_code"] == 200


@pytest.mark.unit
def test_should_execute_successfully_a_empty_controller_with_correlation_id_as_only_input_parameter(
    given_any_petisco, given_any_info_id, given_headers_provider
):

    logger = FakeLogger()

    @controller_handler(
        logger=logger,
        headers_provider=given_headers_provider(given_any_info_id.get_http_headers()),
    )
    def my_controller(headers=None):
        return Success("Hello Petisco")

    http_response = my_controller()

    assert http_response == ({"message": "OK"}, 200)

    first_logging_message = logger.get_logging_messages()[0]
    second_logging_message = logger.get_logging_messages()[1]

    assert first_logging_message == (
        DEBUG,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Processing Request",
            info_id=given_any_info_id,
        ).to_dict(),
    )
    assert second_logging_message == (
        DEBUG,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Result[status: success | value: Hello Petisco]",
            info_id=given_any_info_id,
        ).to_dict(),
    )


@pytest.mark.unit
def test_should_execute_with_a_failure_a_empty_controller_without_input_parameters(
    given_any_petisco, given_any_info_id, given_headers_provider
):

    logger = FakeLogger()

    @controller_handler(
        logger=logger,
        headers_provider=given_headers_provider(given_any_info_id.get_http_headers()),
    )
    def my_controller(headers=None):
        return isFailure

    http_response = my_controller()

    assert http_response == (
        {"error": {"message": "Unknown Error", "type": "HttpError"}},
        500,
    )

    first_logging_message = logger.get_logging_messages()[0]
    second_logging_message = logger.get_logging_messages()[1]

    assert first_logging_message == (
        DEBUG,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Processing Request",
            info_id=given_any_info_id,
        ).to_dict(),
    )

    assert second_logging_message == (
        ERROR,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Result[status: failure | value: Error]",
            info_id=given_any_info_id,
        ).to_dict(),
    )


@pytest.mark.unit
def test_should_execute_with_a_failure_a_empty_controller_with_correlation_id_as_only_input_parameter(
    given_any_petisco, given_any_info_id, given_headers_provider
):

    logger = FakeLogger()

    @controller_handler(
        logger=logger,
        headers_provider=given_headers_provider(given_any_info_id.get_http_headers()),
    )
    def my_controller(headers=None):
        return isFailure

    http_response = my_controller()

    assert http_response == (
        {"error": {"message": "Unknown Error", "type": "HttpError"}},
        500,
    )

    first_logging_message = logger.get_logging_messages()[0]
    second_logging_message = logger.get_logging_messages()[1]

    assert first_logging_message == (
        DEBUG,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Processing Request",
            info_id=given_any_info_id,
        ).to_dict(),
    )
    assert second_logging_message == (
        ERROR,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Result[status: failure | value: Error]",
            info_id=given_any_info_id,
        ).to_dict(),
    )


@pytest.mark.unit
def test_should_execute_successfully_a_empty_controller_without_input_parameters_and_logger(
    given_any_petisco
):
    @controller_handler()
    def my_controller(headers=None):
        return Success("Hello Petisco")

    http_response = my_controller()

    assert http_response == ({"message": "OK"}, 200)


@pytest.mark.unit
def test_should_execute_successfully_a_filtered_object_by_blacklist(
    given_any_petisco, given_any_info_id, given_headers_provider
):

    logger = FakeLogger()

    @controller_handler(
        logger=logger,
        headers_provider=given_headers_provider(given_any_info_id.get_http_headers()),
    )
    def my_controller(headers=None):
        return Success(b"This are bytes")

    http_response = my_controller()

    assert http_response == ({"message": "OK"}, 200)

    first_logging_message = logger.get_logging_messages()[0]
    second_logging_message = logger.get_logging_messages()[1]

    assert first_logging_message == (
        DEBUG,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Processing Request",
            info_id=given_any_info_id,
        ).to_dict(),
    )
    assert second_logging_message == (
        DEBUG,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Success result of type: bytes",
            info_id=given_any_info_id,
        ).to_dict(),
    )


@pytest.mark.unit
def test_should_log_an_exception_occurred_on_the_controller(
    given_any_petisco, given_any_info_id, given_headers_provider
):

    logger = FakeLogger()

    @controller_handler(
        logger=logger,
        headers_provider=given_headers_provider(given_any_info_id.get_http_headers()),
    )
    def my_controller(headers=None):
        raise RuntimeError("my_controller exception")

    http_response = my_controller()

    assert http_response == (
        {"error": {"message": "Internal Error.", "type": "InternalHttpError"}},
        500,
    )

    first_logging_message = logger.get_logging_messages()[0]
    second_logging_message = logger.get_logging_messages()[1]

    assert first_logging_message == (
        DEBUG,
        LogMessageMother.get_controller(
            operation="my_controller",
            message="Processing Request",
            info_id=given_any_info_id,
        ).to_dict(),
    )

    assert second_logging_message[0] == ERROR
    assert "line" in second_logging_message[1]["data"]["message"]
    assert "RuntimeError" in second_logging_message[1]["data"]["message"]
    assert "my_controller exception" in second_logging_message[1]["data"]["message"]


@pytest.mark.unit
def test_should_notify_an_exception_occurred_on_the_controller(
    given_any_petisco, given_any_info_id, given_headers_provider
):

    notifier = FakeNotifier()

    @controller_handler(
        notifier=notifier,
        headers_provider=given_headers_provider(given_any_info_id.get_http_headers()),
    )
    def my_controller(headers=None):
        raise RuntimeError("my_controller exception")

    http_response = my_controller()

    assert http_response == (
        {"error": {"message": "Internal Error.", "type": "InternalHttpError"}},
        500,
    )

    assert notifier.publish_called
    assert notifier.publish_times_called == 1


@pytest.mark.unit
def test_should_notify_a_critical_error_on_the_controller(
    given_any_petisco, given_any_info_id, given_headers_provider
):

    notifier = FakeNotifier()

    class MyCriticalError(CriticalError):
        pass

    @controller_handler(
        notifier=notifier,
        headers_provider=given_headers_provider(given_any_info_id.get_http_headers()),
    )
    def my_controller(headers=None):
        return Failure(MyCriticalError())

    http_response = my_controller()

    assert http_response == (
        {"error": {"message": "Internal Error.", "type": "InternalHttpError"}},
        500,
    )

    assert notifier.publish_called
    assert notifier.publish_times_called == 1



import base64
import os

from petisco.domain.value_objects.uuid import Uuid
from petisco.domain.errors.length_limit_string_value_object_error import (
    ExceedLengthLimitValueObjectError,
)

LENGTH = 16


class UserId(Uuid):
    def guard(self):
        self._ensure_is_16_char_uuid()

    def _ensure_is_16_char_uuid(self):
        if len(self.value) > LENGTH:
            raise ExceedLengthLimitValueObjectError(message=self.value)

    @classmethod
    def generate(cls):
        # TODO: refactor to cls(str(uuid.uuid4()))
        r = os.urandom(LENGTH)
        return cls(base64.b64encode(r, altchars=b"-_").decode("utf-8")[:LENGTH])



import pytest

from petisco.domain.aggregate_roots.aggregate_root import AggregateRoot
from petisco.domain.value_objects.name import Name
from petisco.domain.value_objects.user_id import UserId
from petisco.event.shared.domain.event import Event


class UserCreated(Event):
    user_id: UserId
    name: Name

    def __init__(self, user_id: UserId, name: Name):
        self.user_id = user_id
        self.name = name
        super().__init__()


class User(AggregateRoot):
    def __init__(self, name: Name, user_id: UserId):
        self.name = name
        self.user_id = user_id
        super().__init__()

    @staticmethod
    def create(name: Name):
        user = User(name, UserId.generate())
        user.record(UserCreated(user.user_id, user.name))
        return user


@pytest.mark.unit
def test_should_play_with_player_aggregate_root():

    user = User.create(Name("Petisco"))

    assert isinstance(user.pull_first_domain_event(), UserCreated)
    assert isinstance(user.pull_last_domain_event(), UserCreated)
    assert len(user.pull_domain_events()) == 1

    user.clear_domain_events()

    assert user.pull_first_domain_event() is None
    assert user.pull_last_domain_event() is None
    assert len(user.pull_domain_events()) == 0



from typing import Dict
from abc import ABCMeta, abstractmethod


from petisco.event.shared.domain.event import Event, Events


class IEventBus:

    __metaclass__ = ABCMeta

    def __repr__(self):
        return f"IEventPublisher"

    @abstractmethod
    def info(self) -> Dict:
        raise NotImplementedError

    @abstractmethod
    def publish(self, event: Event):
        raise NotImplementedError

    def publish_events(self, events: Events):
        for event in events:
            self.publish(event)

    @abstractmethod
    def stop(self):
        raise NotImplementedError



from petisco.event.shared.domain.config_events import ConfigEvents
from petisco.event.bus.infrastructure.not_implemented_event_bus import (
    NotImplementedEventBus,
)
from petisco.event.configurer.infrastructure.not_implemented_configurer import (
    NotImplementedEventConfigurer,
)
from petisco.event.consumer.infrastructure.not_implemented_event_comsumer import (
    NotImplementedEventConsumer,
)


def configure_events_infrastructure(config_events: ConfigEvents):
    bus = NotImplementedEventBus()
    configurer = NotImplementedEventConfigurer()
    consumer = NotImplementedEventConsumer()

    if config_events.message_broker == "rabbitmq":
        from petisco.event.shared.infrastructure.rabbitmq.rabbitmq_connector import (
            RabbitMqConnector,
        )
        from petisco.event.bus.infrastructure.rabbitmq_event_bus import RabbitMqEventBus
        from petisco.event.configurer.infrastructure.rabbitmq_configurer import (
            RabbitMqEventConfigurer,
        )
        from petisco.event.consumer.infrastructure.rabbitmq_event_comsumer import (
            RabbitMqEventConsumer,
        )

        # connector = RabbitMqConnector()
        bus = RabbitMqEventBus(
            connector=RabbitMqConnector(),
            organization=config_events.organization,
            service=config_events.service,
        )
        configurer = RabbitMqEventConfigurer(
            connector=RabbitMqConnector(),
            organization=config_events.organization,
            service=config_events.service,
            use_store_queues=config_events.use_store_queues,
            retry_ttl=config_events.retry_ttl,
        )
        consumer = RabbitMqEventConsumer(
            connector=RabbitMqConnector(),
            organization=config_events.organization,
            service=config_events.service,
            max_retries=config_events.max_retries,
            verbose=config_events.consumer_verbose,
        )

    return bus, configurer, consumer



import inspect

import traceback

from petisco.application.petisco import Petisco
from petisco.domain.errors.critical_error import CriticalError
from petisco.domain.errors.unknown_error import UnknownError
from petisco.logger.interface_logger import DEBUG
from petisco.event.shared.domain.event import Event
from functools import wraps
from meiga import Result, Failure
from meiga.decorators import meiga

from petisco.logger.log_message import LogMessage
from petisco.notifier.domain.notifier_exception_message import NotifierExceptionMessage
from petisco.notifier.infrastructure.not_implemented_notifier import (
    NotImplementedNotifier,
)

DEFAULT_LOGGER = None
DEFAULT_NOTIFIER = None


class _EventHandler:
    def __init__(self, logger=DEFAULT_LOGGER, notifier=DEFAULT_NOTIFIER):
        """
        Parameters
        ----------
        logger
            A ILogger implementation. Default NotImplementedLogger
        notifier
            A INotifier implementation. If not specified it will get it from Petisco.get_notifier(). You can also use NotImplementedNotifier
        """
        self.logger = logger
        self.notifier = notifier

    def _check_logger(self):
        if self.logger == DEFAULT_LOGGER:
            from petisco import Petisco

            self.logger = Petisco.get_logger()

    def _check_notifier(self):
        if self.notifier == DEFAULT_NOTIFIER:
            self.notifier = NotImplementedNotifier()

    def __call__(self, func, *args, **kwargs):
        @wraps(func)
        def wrapper(*args, **kwargs):
            @meiga
            def run_event_handler(**kwargs) -> Result:
                params = inspect.getfullargspec(func).args
                kwargs = {k: v for k, v in kwargs.items() if k in params}
                return func(**kwargs)

            self._check_logger()
            self._check_notifier()

            event: Event = args[0]

            log_message = LogMessage(
                layer="event_handler", operation=f"{func.__name__}"
            )

            self.logger.log(
                DEBUG,
                log_message.set_message(
                    {"event": event.event_name, "body": event.to_json()}
                ),
            )

            kwargs = dict(event=event)

            try:
                result = run_event_handler(**kwargs)
            except Exception as exception:
                result = Failure(
                    UnknownError(
                        exception=exception,
                        input_parameters=kwargs if len(kwargs) > 0 else args,
                        executor=f"{func.__name__} (Event Handler)",
                        traceback=traceback.format_exc(),
                    )
                )

            self.notify(result)

        return wrapper

    @meiga
    def notify(self, result):
        if result.is_failure:
            error = result.value
            if issubclass(error.__class__, CriticalError):
                self.notifier.publish(
                    NotifierExceptionMessage(
                        exception=error.exception,
                        executor=error.executor,
                        input_parameters=error.input_parameters,
                        traceback=error.traceback,
                        info_petisco=Petisco.get_info(),
                    )
                )


event_handler = _EventHandler



from typing import List, Callable

from dataclasses import dataclass


@dataclass
class EventSubscriber:
    event_name: str
    event_version: int
    handlers: List[Callable]

    def get_handlers_names(self) -> List[str]:
        return [handler.__name__ for handler in self.handlers]



from typing import Dict

from meiga import Result, Error, NotImplementedMethodError

from petisco.application.interface_repository import IRepository
from petisco.domain.value_objects.client_id import ClientId
from petisco.domain.value_objects.user_id import UserId
from tests.end2end.flask_app.user.domain.aggregate_roots.user import User


class IUserRepository(IRepository):
    def info(self) -> Dict:
        return {"name": self.__class__.__name__}

    def save(self, user: User) -> Result[bool, Error]:
        return NotImplementedMethodError

    def retrieve(self, client_id: ClientId, user_id: UserId) -> Result[User, Error]:
        return NotImplementedMethodError



from typing import Callable, List

from petisco.event.shared.domain.event_subscriber import EventSubscriber
from petisco.event.consumer.domain.interface_event_consumer import IEventConsumer


class NotImplementedEventConsumer(IEventConsumer):
    def start(self):
        pass

    def add_subscribers(self, subscribers: List[EventSubscriber]):
        pass

    def add_subscriber_on_dead_letter(
        self, subscriber: EventSubscriber, handler: Callable
    ):
        pass

    def add_handler_on_store(self, handler: Callable):
        pass

    def add_handler_on_queue(self, queue_name: str, handler: Callable):
        pass

    def stop(self):
        pass



from petisco import RabbitMqConnector
from petisco.event.consumer.infrastructure.rabbitmq_event_comsumer import (
    RabbitMqEventConsumer,
)
from tests.modules.event.mothers.defaults import (
    DEFAULT_ORGANIZATION,
    DEFAULT_SERVICE,
    DEFAULT_MAX_RETRIES,
    DEFAULT_VERBOSE,
)


class RabbitMqEventConsumerMother:
    @staticmethod
    def default(connector: RabbitMqConnector = None):
        connector = RabbitMqConnector() if not connector else connector
        return RabbitMqEventConsumer(
            connector,
            DEFAULT_ORGANIZATION,
            DEFAULT_SERVICE,
            DEFAULT_MAX_RETRIES,
            DEFAULT_VERBOSE,
        )

    @staticmethod
    def with_max_retries(max_retries: int, connector: RabbitMqConnector = None):
        connector = RabbitMqConnector() if not connector else connector
        return RabbitMqEventConsumer(
            connector,
            DEFAULT_ORGANIZATION,
            DEFAULT_SERVICE,
            max_retries,
            DEFAULT_VERBOSE,
        )

    @staticmethod
    def without_retry(connector: RabbitMqConnector = None):
        connector = RabbitMqConnector() if not connector else connector
        return RabbitMqEventConsumer(
            connector, DEFAULT_ORGANIZATION, DEFAULT_SERVICE, 0, DEFAULT_VERBOSE
        )



import pytest
import requests
import requests_mock
from meiga.assertions import assert_success, assert_failure

from petisco import (
    Request,
    MissingSchemaRequestError,
    TimeoutRequestError,
    ConnectionRequestError,
)


@pytest.fixture
def given_any_url():
    return "http://alicebiometrics.com"


@pytest.fixture
def given_invalid_url():
    return "invalid_url"


@pytest.mark.integration
def test_should_success_when_request_get(given_any_url):
    with requests_mock.Mocker() as m:
        m.get(given_any_url)
        result = Request.get(given_any_url)
        assert_success(result)


@pytest.mark.integration
def test_should_success_when_request_post(given_any_url):
    with requests_mock.Mocker() as m:
        m.post(given_any_url)
        result = Request.post(given_any_url)
        assert_success(result)


@pytest.mark.integration
def test_should_success_when_request_patch(given_any_url):
    with requests_mock.Mocker() as m:
        m.patch(given_any_url)
        result = Request.patch(given_any_url)
        assert_success(result)


@pytest.mark.integration
def test_should_success_when_request_delete(given_any_url):
    with requests_mock.Mocker() as m:
        m.delete(given_any_url)
        result = Request.delete(given_any_url)
        assert_success(result)


@pytest.mark.integration
@pytest.mark.parametrize("method", ["GET", "POST", "PATCH", "DELETE"])
def test_should_fail_when_request_method_raise_missing_schema_error(
    method, given_any_url
):
    with requests_mock.Mocker() as m:
        m.register_uri(method, given_any_url, exc=requests.exceptions.MissingSchema)
        result = Request.execute(given_any_url, method)
        assert_failure(result, value_is_instance_of=MissingSchemaRequestError)


@pytest.mark.integration
@pytest.mark.parametrize("method", ["GET", "POST", "PATCH", "DELETE"])
def test_should_fail_when_request_method_raise_timeout_error(method, given_any_url):
    with requests_mock.Mocker() as m:
        m.register_uri(method, given_any_url, exc=requests.exceptions.ConnectTimeout)
        result = Request.execute(given_any_url, method)
        assert_failure(result, value_is_instance_of=TimeoutRequestError)


@pytest.mark.integration
@pytest.mark.parametrize("method", ["GET", "POST", "PATCH", "DELETE"])
def test_should_fail_when_request_method_raise_connection_error(method, given_any_url):
    with requests_mock.Mocker() as m:
        m.register_uri(method, given_any_url, exc=requests.exceptions.ConnectionError)
        result = Request.execute(given_any_url, method)
        assert_failure(result, value_is_instance_of=ConnectionRequestError)



from petisco import Event, EventId


class SpyEvents:
    def __init__(self):
        self.counter_events = {}
        self.events = []

    def append(self, event: Event):
        self.events.append(event)
        if str(event.event_id) not in self.counter_events:
            self.counter_events[str(event.event_id)] = {"counter": 1}
        else:
            self.counter_events[str(event.event_id)]["counter"] += 1

    def assert_number_unique_events(self, expected_number_events: int):
        actual_number_events = len(self.counter_events.keys())
        assert (
            actual_number_events == expected_number_events
        ), f"Expected events is {expected_number_events}, actual {actual_number_events}. [{self.counter_events}]"

    def get_counter_by_event_id(self, event_id: EventId):
        return self.counter_events.get(str(event_id), {}).get("counter", 0)

    def assert_count_by_event_id(self, event_id: EventId, expected_count: int):
        assert (
            self.get_counter_by_event_id(event_id) == expected_count
        ), f"Expected events is {expected_count}, actual {self.get_counter_by_event_id(event_id)}. [{self.counter_events}]"

    def assert_first_event(self, expected_event: Event):
        self.assert_number_unique_events(1)
        event = self.events[0]
        assert event == expected_event

    def assert_last_event(self, expected_event: Event):
        self.assert_number_unique_events(1)
        event = self.events[-1]
        assert event == expected_event



from pymongo import MongoClient
from pymongo.errors import ConnectionFailure


def mongodb_is_running_locally(
    host="localhost", port=27017, username="root", password="password123"
) -> bool:
    mongo_client = MongoClient(
        f"mongodb://{host}:{port}/",
        username=username,
        password=password,
        serverSelectionTimeoutMS=1000,
    )
    try:
        # The ismaster command is cheap and does not require auth.
        mongo_client.admin.command("ismaster")
    except ConnectionFailure:
        return False
    return True



from meiga import Result, Error

from petisco import Event


def send_mail_handler(event: Event) -> Result[bool, Error]:
    pass


def send_sms_handler(event: Event) -> Result[bool, Error]:
    pass



from typing import List

from petisco.event.shared.domain.event import Event
from petisco.event.shared.domain.event_subscriber import EventSubscriber
from petisco.event.configurer.domain.interface_event_configurer import IEventConfigurer
from petisco.event.shared.infrastructure.rabbitmq.rabbitmq_exchange_name_formatter import (
    RabbitMqExchangeNameFormatter,
)
from petisco.event.shared.infrastructure.rabbitmq.rabbitmq_event_subscriber_queue_name_formatter import (
    RabbitMqEventSubscriberQueueNameFormatter,
)
from petisco.event.shared.infrastructure.rabbitmq.rabbitmq_connector import (
    RabbitMqConnector,
)


class RabbitMqEventConfigurer(IEventConfigurer):
    def __init__(
        self,
        connector: RabbitMqConnector,
        organization: str,
        service: str,
        use_store_queues: bool = True,
        retry_ttl: int = 5000,
    ):
        self._connector = connector
        self._exchange_name = f"{organization}.{service}"
        self._use_store_queues = use_store_queues
        self._configured_subscribers = []
        self.retry_ttl = retry_ttl

    def configure(self):
        self.configure_subscribers([])

    def configure_event(self, event: Event):
        self.configure_subscribers(
            [
                EventSubscriber(
                    event_name=event.event_name,
                    event_version=event.event_version,
                    handlers=[],
                )
            ]
        )

    def configure_subscribers(self, subscribers: List[EventSubscriber]):
        self._configure_exchanges()
        self._declare_queues(
            self._exchange_name,
            self._retry_exchange_name,
            self._dead_letter_exchange_name,
            subscribers,
        )
        self._configured_subscribers.append(subscribers)

    def clear(self):
        self._delete_exchange()
        self._delete_queues()

    def _delete_exchange(self):
        channel = self._connector.get_channel(self._exchange_name)
        channel.exchange_delete(self._exchange_name)
        channel.exchange_delete(self._retry_exchange_name)
        channel.exchange_delete(self._dead_letter_exchange_name)

    def _delete_queues(self):
        channel = self._connector.get_channel(self._exchange_name)
        channel.queue_delete("store")
        channel.queue_delete("retry.store")
        channel.queue_delete("dead_letter.store")
        for subscribers in self._configured_subscribers:
            for subscriber in subscribers:
                queue_name = RabbitMqEventSubscriberQueueNameFormatter.format(
                    subscriber, exchange_name=self._exchange_name
                )
                retry_queue_name = RabbitMqEventSubscriberQueueNameFormatter.format_retry(
                    subscriber, exchange_name=self._exchange_name
                )
                dead_letter_queue_name = RabbitMqEventSubscriberQueueNameFormatter.format_dead_letter(
                    subscriber, exchange_name=self._exchange_name
                )
                for suffix in subscriber.get_handlers_names():
                    name = f"{queue_name}.{suffix}"
                    retry_name = f"{retry_queue_name}.{suffix}"
                    dead_letter_name = f"{dead_letter_queue_name}.{suffix}"
                    channel.queue_delete(name)
                    channel.queue_delete(retry_name)
                    channel.queue_delete(dead_letter_name)

    def _configure_exchanges(self):
        self._retry_exchange_name = RabbitMqExchangeNameFormatter.retry(
            self._exchange_name
        )
        self._dead_letter_exchange_name = RabbitMqExchangeNameFormatter.dead_letter(
            self._exchange_name
        )
        self._declare_exchange(self._exchange_name)
        self._declare_exchange(self._retry_exchange_name)
        self._declare_exchange(self._dead_letter_exchange_name)

    def _declare_exchange(self, exchange_name: str):
        channel = self._connector.get_channel(self._exchange_name)
        try:
            channel.exchange_declare(
                exchange=exchange_name, exchange_type="topic", durable=True
            )
        except Exception as error:
            raise TypeError(
                f"RabbitMQEventPublisher: Cannot create the exchange ({exchange_name})\n{error}"
            )

    def _declare_queues(
        self,
        exchange_name: str,
        retry_exchange_name: str,
        dead_letter_exchange_name: str,
        subscribers: List[EventSubscriber],
    ):
        if self._use_store_queues:
            self._declare_store_queues(
                exchange_name, retry_exchange_name, dead_letter_exchange_name
            )

        for subscriber in subscribers:
            base_queue_name = RabbitMqEventSubscriberQueueNameFormatter.format(
                subscriber, exchange_name=exchange_name
            )
            base_retry_queue_name = RabbitMqEventSubscriberQueueNameFormatter.format_retry(
                subscriber, exchange_name=exchange_name
            )
            base_dead_letter_queue_name = RabbitMqEventSubscriberQueueNameFormatter.format_dead_letter(
                subscriber, exchange_name=exchange_name
            )
            routing_key = base_queue_name

            for suffix in subscriber.get_handlers_names():
                queue_name = f"{base_queue_name}.{suffix}"
                retry_queue_name = f"{base_retry_queue_name}.{suffix}"
                dead_letter_queue_name = f"{base_dead_letter_queue_name}.{suffix}"

                self._declare_queue(queue_name=queue_name)
                self._declare_queue(
                    queue_name=retry_queue_name,
                    dead_letter_exchange=exchange_name,
                    dead_letter_routing_key=f"retry.{queue_name}",
                    message_ttl=self.retry_ttl,
                )
                self._declare_queue(queue_name=dead_letter_queue_name)

                self._bind_queue(
                    exchange_name=exchange_name,
                    queue_name=queue_name,
                    routing_key=routing_key,
                )
                self._bind_queue(
                    exchange_name=exchange_name,
                    queue_name=queue_name,
                    routing_key=f"retry.{queue_name}",
                )
                self._bind_queue(
                    exchange_name=retry_exchange_name,
                    queue_name=retry_queue_name,
                    routing_key=f"retry.{queue_name}",
                )
                self._bind_queue(
                    exchange_name=dead_letter_exchange_name,
                    queue_name=dead_letter_queue_name,
                    routing_key=f"dead_letter.{queue_name}",
                )

    def _declare_store_queues(
        self,
        exchange_name: str,
        retry_exchange_name: str,
        dead_letter_exchange_name: str,
    ):
        self._declare_queue(queue_name="store")
        self._declare_queue(
            queue_name="retry.store",
            dead_letter_exchange=exchange_name,
            dead_letter_routing_key="retry.store",
            message_ttl=self.retry_ttl,
        )
        self._declare_queue(queue_name="dead_letter.store")

        routing_key_any_event = f"*.*.*.event.*"
        self._bind_queue(
            exchange_name=exchange_name,
            queue_name="store",
            routing_key=routing_key_any_event,
        )
        self._bind_queue(
            exchange_name=exchange_name, queue_name="store", routing_key="retry.store"
        )
        self._bind_queue(
            exchange_name=retry_exchange_name,
            queue_name="retry.store",
            routing_key="retry.store",
        )
        self._bind_queue(
            exchange_name=retry_exchange_name,
            queue_name="retry.store",
            routing_key=f"retry.store",
        )
        self._bind_queue(
            exchange_name=dead_letter_exchange_name,
            queue_name="dead_letter.store",
            routing_key=f"dead_letter.{routing_key_any_event}",
        )

    def _declare_queue(
        self,
        queue_name: str,
        dead_letter_exchange: str = None,
        dead_letter_routing_key: str = None,
        message_ttl: int = None,
    ):
        channel = self._connector.get_channel(self._exchange_name)

        queue_arguments = {}
        if dead_letter_exchange:
            queue_arguments["x-dead-letter-exchange"] = dead_letter_exchange

        if dead_letter_routing_key:
            queue_arguments["x-dead-letter-routing-key"] = dead_letter_routing_key

        if message_ttl:
            queue_arguments["x-message-ttl"] = message_ttl

        result = channel.queue_declare(
            queue=queue_name, arguments=queue_arguments, durable=True
        )

        return result.method.queue

    def _bind_queue(self, exchange_name: str, queue_name: str, routing_key: str):
        channel = self._connector.get_channel(self._exchange_name)

        channel.queue_bind(
            exchange=exchange_name, queue=queue_name, routing_key=routing_key
        )



from meiga import Result, Error

from petisco import Event


def send_mail_on_user_created(event: Event) -> Result[bool, Error]:
    pass


def send_referral_on_user_created(event: Event) -> Result[bool, Error]:
    pass


def send_promotion_on_user_created(event: Event) -> Result[bool, Error]:
    pass



from typing import Dict

from meiga import Result, Error, isSuccess, Success, Failure
from pymongo import MongoClient
from pymongo.database import Database

from petisco import Petisco
from petisco.domain.value_objects.client_id import ClientId
from petisco.domain.value_objects.name import Name
from petisco.domain.value_objects.user_id import UserId
from tests.end2end.flask_app.shared.domain.repositories.interface_user_repository import (
    IUserRepository,
)
from tests.end2end.flask_app.shared.domain.repositories.user_already_exist_error import (
    UserAlreadyExistError,
)
from tests.end2end.flask_app.shared.domain.repositories.user_not_found_error import (
    UserNotFoundError,
)
from tests.end2end.flask_app.user.domain.aggregate_roots.user import User


class MongoDBUserRepository(IUserRepository):
    @staticmethod
    def build():
        return MongoDBUserRepository(
            client=Petisco.persistence_mongodb_client(),
            database=Petisco.persistence_mongodb_database(),
        )

    def __init__(self, client: MongoClient, database: Database):
        self.client = client
        self.database = database
        self.collection = self.database["user"]

    def info(self) -> Dict:
        return {"name": self.__class__.__name__}

    def save(self, user: User) -> Result[bool, Error]:
        result = self.exists(user.user_id)
        if result.is_success:
            return Failure(UserAlreadyExistError(user.user_id))

        if self.collection.insert_one(user.to_dict()):
            return isSuccess

    def retrieve(self, client_id: ClientId, user_id: UserId) -> Result[User, Error]:
        user_doc = self.collection.find_one(
            {"user_id": user_id.value, "client_id": client_id.value}
        )
        if user_doc:
            return Success(
                User(
                    name=Name(user_doc.name),
                    client_id=ClientId(user_doc.client_id),
                    user_id=UserId(user_doc.user_id),
                )
            )
        else:
            return Failure(UserNotFoundError(user_id))

    def exists(self, user_id: UserId) -> Result[bool, Error]:
        if self.collection.find_one({"user_id": user_id.value}):
            return isSuccess
        else:
            return Failure(UserNotFoundError(user_id))



import importlib
import sys
from typing import Optional, Dict, Any, Callable, List

from dataclasses import dataclass

from petisco.application.config.get_funtion_from_string import get_function_from_string
from petisco.application.config.raise_petisco_config_error import (
    raise_petisco_config_exception,
)


@dataclass
class SingleConfigPersistence:
    type: str
    config: Optional[Callable] = None
    models: Optional[Dict[str, Any]] = None


@dataclass
class ConfigPersistence:

    configs: Optional[Dict[str, SingleConfigPersistence]] = None

    @staticmethod
    def from_dict(kdict):
        if not kdict or not isinstance(kdict, List):
            return ConfigPersistence()
        configs = {}
        for entry in kdict:
            config = (
                get_function_from_string(entry.get("config"))
                .handle(
                    on_failure=raise_petisco_config_exception,
                    failure_args=(entry, "persistence:config"),
                )
                .unwrap()
            )
            configs[entry["name"]] = SingleConfigPersistence(
                type=entry["type"], config=config, models=entry.get("models")
            )
        return ConfigPersistence(configs=configs)

    def get_models(self, persistence_entry):
        loaded_models = {}
        for name, model_string in self.configs[persistence_entry].models.items():
            mod_name, model_name = model_string.rsplit(".", 1)
            mod = importlib.import_module(mod_name)
            loaded_models[name] = getattr(mod, model_name)
        return loaded_models

    def get_import_database_models_func(self, persistence_entry):
        def _delete_module_if_already_imported(module_name: str):
            # Important. SqlAlchemy needs models to be imported
            # Base = SqlAlchemyPersistence.get_instance().sources["petisco"]["base"]
            # If module is already imported and Base has not tables, Tables won't be imported.
            # e.g len(Base.metadata.tables) == 0
            # With this function we ensure that model can be imported to help SqlAlchemy loading models
            if module_name in sys.modules:
                del sys.modules[module_name]

        def _import_database_models_func():
            for name, model_string in self.configs[persistence_entry].models.items():
                module_name, class_model_name = model_string.rsplit(".", 1)
                _delete_module_if_already_imported(module_name)
                importlib.import_module(module_name, class_model_name)

        return _import_database_models_func



class RabbitMqExchangeNameFormatter:
    @staticmethod
    def retry(exchange_name: str) -> str:
        return f"retry.{exchange_name}"

    @staticmethod
    def dead_letter(exchange_name: str) -> str:
        return f"dead_letter.{exchange_name}"



from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL

import pytest
import os

from petisco import Petisco


@pytest.mark.end2end
def test_should_load_petisco_from_yml(petisco_yml_path_end2end, given_petisco_version):

    filename = f"{petisco_yml_path_end2end}/petisco.all.yml"

    expected_petisco_info = {
        "app_name": "toy-app",
        "app_version": "1.0.0",
        "petisco_version": given_petisco_version,
        "environment": None,
        "services": {"sum": {"name": "SumExecutor"}},
        "repositories": {
            "user": {"name": "SqlUserRepository"},
            "users_count": {"name": "SqlUsersCountRepository"},
        },
        "event_publisher": {"name": "NotImplementedEventPublisher"},
        "event_subscriber": {"name": "NotImplementedEventSubscriber"},
        "tasks": {
            "recurring-task": {"type": "recurring", "run_in": 0.0, "interval": 10.0},
            "scheduled-task": {"type": "scheduled", "run_in": 10.0},
            "instant-task": {"type": "instant", "run_in": 0.0},
        },
    }

    petisco = Petisco.from_filename(filename)
    petisco.load_services_and_repositories()

    assert "elapsed_time" in petisco.info
    assert "load_repositories" in petisco.info["elapsed_time"]
    assert "load_services" in petisco.info["elapsed_time"]
    petisco.info.pop("elapsed_time")
    assert petisco.info == expected_petisco_info
    Petisco.clear()


@pytest.mark.end2end
def test_should_load_petisco_from_yml_and_schedule_tasks(petisco_yml_path_end2end):

    filename = f"{petisco_yml_path_end2end}/petisco.all.yml"

    petisco = Petisco.from_filename(filename)

    petisco._schedule_tasks()

    Petisco.clear()


@pytest.mark.end2end
@pytest.mark.parametrize(
    "logging_level, logging_level_num",
    [
        ("DEBUG", DEBUG),
        ("INFO", INFO),
        ("WARNING", WARNING),
        ("ERROR", ERROR),
        ("CRITICAL", CRITICAL),
    ],
)
def test_should_load_petisco_from_yml_and_check_logger_level(
    logging_level, logging_level_num, petisco_yml_path_end2end
):

    Petisco.clear()
    os.environ["PETISCO_LOGGING_LEVEL"] = logging_level

    filename = f"{petisco_yml_path_end2end}/petisco.all.withlogging.yml"

    petisco = Petisco.from_filename(filename)

    assert hasattr(petisco.get_logger(), "logging_level")
    assert petisco.get_logger().logging_level == logging_level
    assert petisco.get_logger().logger.level == logging_level_num

    del os.environ["PETISCO_LOGGING_LEVEL"]
    Petisco.clear()


@pytest.mark.end2end
def test_should_raise_exception_if_repository_not_exist(petisco_yml_path_end2end):

    filename = f"{petisco_yml_path_end2end}/petisco.all.yml"

    petisco = Petisco.from_filename(filename)
    with pytest.raises(ValueError):
        petisco.get_repository("repo")

    Petisco.clear()


@pytest.mark.end2end
def test_should_raise_exception_if_service_not_exist(petisco_yml_path_end2end):

    filename = f"{petisco_yml_path_end2end}/petisco.all.yml"

    petisco = Petisco.from_filename(filename)
    with pytest.raises(ValueError):
        petisco.get_service("repo")

    Petisco.clear()


@pytest.mark.end2end
def test_should_check_if_given_any_petisco_fixture_is_working_as_expected(
    given_any_petisco
):
    assert given_any_petisco.app_name == "petisco"
    assert given_any_petisco.app_version == "1.0.0"


@pytest.mark.end2end
def test_should_check_error_when_petisco_is_not_initialized():
    with pytest.raises(Exception) as excinfo:
        Petisco.get_instance()
        solutions_text = (
            f"\t Possible Solutions:\n"
            f"\t  * If you are testing and need any preconfigured Petisco, you can use the fixture given_any_petisco.\n"
            f"\t  * If you are testing and need specific preconfigured Petisco, you can create a fixture loading a petisco and then clearing.\n"
            f"\t  * If you are running a Petisco application, please review your loading process, petisco.yml is probably failing.\n"
        )
        assert solutions_text in str(excinfo.value)
